"""Object-orientd and cross platform file system path management."""


import abc
import pathlib
from typing import Any, IO, Iterator, Optional, Protocol, Union, runtime_checkable


@runtime_checkable
class Path(Protocol):
    """Abstract base class for specifying cross platform path interfaces."""

    # @abc.abstractmethod
    # def __enter__(self):

    #     raise NotImplementedError

    # @abc.abstractmethod
    # def __exit__(self):

    #     raise NotImplementedError

    @abc.abstractmethod
    def exists(self) -> bool:
        """Whether this path exists.

        Return:
            whether this path exists
        """

        raise NotImplementedError

    @abc.abstractmethod
    def is_dir(self) -> bool:
        """Whether this path is a directory.

        Return:
            whether this path is a directory
        """

        raise NotImplementedError

    @abc.abstractmethod
    def is_file(self) -> bool:
        """Whether this path is a regular file.

        Return:
            whether this path is a regular file
        """

        raise NotImplementedError

    @abc.abstractmethod
    def iterdir(self) -> Iterator["Path"]:
        """Iterate over the files in this directory.

        Does not yield any result for the special paths '.' and '..'.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def glob(self, pattern: str) -> Iterator["Path"]:
        """Iterate over this subtree and yield all existing files matching the given relative pattern.

        Args:
            pattern: Unix glob pattern for file matching

        Yield:
            paths matching pattern including directories
        """

        raise NotImplementedError

    @abc.abstractmethod
    def mkdir(self) -> None:
        """Create a new directory at this given path."""

        raise NotImplementedError

    @abc.abstractmethod
    def open(self, mode: str = "r", buffering: int = -1, encoding: Any = None) -> IO:
        """Open the file pointed by this path and return a file object.

        Args:
            mode:
            buffering:
            encoding: encoding used to decode or encode the file in text mode

        Return:
            handle to file object
        """

        raise NotImplementedError

    @abc.abstractmethod
    def read_bytes(self) -> bytes:
        """Open the file in bytes mode, read it, and close the file.

        Return:
            bytes data from file
        """

        raise NotImplementedError

    @abc.abstractmethod
    def read_text(self, encoding: Any = None, errors: Any = None) -> str:
        """Open the file in text mode, read it, and close the file.

        Args:
            encoding: encoding used to decode the file
            errors:

        Return:
            text data from file
        """

        raise NotImplementedError

    @abc.abstractmethod
    def rename(self, target: Union[str, pathlib.PurePath]) -> None:
        """Rename this path to the given path.

        Args:
            target: rename target Path
        """

        raise NotImplementedError

    @abc.abstractmethod
    def replace(self, target: Union[str, pathlib.PurePath]) -> None:
        """Rename this path to the given path and overwrite the existing destination if it exists.

        Args:
            target: replace target Path

        Return:
            new Path instance pointing to the target path
        """

        raise NotImplementedError

    @abc.abstractmethod
    def rglob(self, pattern: str) -> Iterator["Path"]:
        """Recursively yield all existing files matching the given relative pattern.

        Args:
            pattern:
        """

        raise NotImplementedError

    @abc.abstractmethod
    def rmdir(self) -> None:
        """Remove this directory if empty."""

        raise NotImplementedError

    @abc.abstractmethod
    def touch(self) -> None:
        """Create this file if it doesn't exist."""

        raise NotImplementedError

    @abc.abstractmethod
    def write_bytes(self, data: bytes) -> int:
        """Open the file in bytes mode, write to it, and close the file.

        Args:
            data:
        """

        raise NotImplementedError

    @abc.abstractmethod
    def write_text(
        self, data: str, encoding: Optional[str] = None, errors: Optional[str] = None
    ) -> int:
        """Open the file in text mode, write to it, and close the file.

        Args:
            data:
            encoding:
            errors:
        """

        raise NotImplementedError


Path.register(pathlib.Path)
