"""Cloud storage pathlib interface."""


import abc
import functools
import os
import pathlib
from typing import Any, Union

from cloudpath.base import base


@functools.total_ordering
class CloudPath(abc.ABC):
    """Cloud storage pathlib interface."""

    def __init__(self, path: Union[str, pathlib.Path]):
        """Create a new PathS3 instance.

        Raises:
            ValueError: if path is not a valid S3 path
        """

        path_name = str(path)
        if self.valid_path(path_name):
            self._path = pathlib.Path(path_name)
        else:
            raise ValueError(f"{path_name} is not a valid S3 path")

        self._bucket_name = self._path.parts[1]
        self._load_resource()

    def as_uri(self) -> str:
        """Format S3 path as valid URI.

        Return:
            valid URI path
        """

        return f"{self.scheme()}://{str(self._path)[1:]}"

    @property
    def _cache_path(self) -> pathlib.Path:
        """Find local file path for caching S3 file.

        Return:
            local file path
        """

        return self._cache_dir() / str(self._path)[1:]

    @classmethod
    def _cache_dir(cls) -> pathlib.Path:
        """Find local storage cache directory and create it if nonexistant.

        Return:
            local cache directory for class
        """

        cache_dir = base.cache_dir() / cls.scheme()
        if not cache_dir.exists():
            cache_dir.mkdir()

        return cache_dir

    def _create_parent(self) -> None:
        """Create parent directory if it does not exist."""

        if not self._cache_path.parent.exists():
            self._cache_path.parent.mkdir()

    @abc.abstractmethod
    def _download_file(self, local_path: str, remote_path: str) -> bool:
        """Download file from remote storage bucket.

        Args:
            local_path: local system file path
            remote_path: remote storage file path

        Return:
            whether file download was successful
        """

        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        """Compare if instance is equal to the other instance.

        Args:
            other: other CloudPath instance

        Return:
            whether instance is equal to other
        """

        return bool(self._path == other._path)

    @abc.abstractmethod
    def _load_bucket(self) -> None:
        """Find and set cloud storage bucket.

        Raises:
            FileNotFoundError: if unable to find storage bucket
        """

        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _load_resource(cls) -> None:
        """Import the library resource for interacting with storage buckets."""

        raise NotImplementedError

    def __lt__(self, other: Any) -> bool:
        """Compare if instance is less than other instance.

        Args:
            other: other CloudPath instance

        Return:
            whether instance is less than other
        """

        return bool(self._path < other._path)

    def __repr__(self) -> str:
        """Create string representation of path.

        Return:
            string representation of path
        """

        return f"{self.__class__.__name__}('{self._path}')"

    def read_text(self, encoding: Any = None, errors: Any = None) -> str:
        """Open the file in text mode, read it, and close the file.

        Args:
            encoding: encoding used to decode the file
            errors:

        Raises:
            IOError: if unable to download file from cloud storage

        Return:
            text data from file
        """

        self._load_bucket()
        self._create_parent()

        key_path = os.path.join(*self._path.parts[2:])
        status = self._download_file(str(self._cache_path), key_path)

        if not status:
            raise IOError(f"Unable to download {self._path} from {self.scheme()}")

        return self._cache_path.read_text(encoding, errors)

    @classmethod
    @abc.abstractmethod
    def scheme(cls) -> str:
        """Return cloud storage URI scheme.

        Return:
            URI scheme
        """

        raise NotImplementedError

    @abc.abstractmethod
    def _upload_file(self, local_path: str, remote_path: str) -> bool:
        """Upload file to remote storage bucket.

        Args:
            local_path: local system file path
            remote_path: remote storage file path

        Return:
            whether file upload was successful
        """

        raise NotImplementedError

    def write_text(self, data: str) -> None:
        """Write text to cloud storage bucket file.

        Args:
            data: text to write

        Raises:
            IOError: if unable to upload file to cloud storage
        """

        self._load_bucket()
        self._create_parent()
        self._cache_path.write_text(data)

        key_path = os.path.join(*self._path.parts[2:])
        status = self._upload_file(str(self._cache_path), key_path)

        if not status:
            raise IOError(f"unable to upload {self._path} to {self.scheme()}")

    @classmethod
    @abc.abstractmethod
    def valid_path(cls, path_name: str) -> bool:
        """Check if cloud path is valid.

        Args:
            full cloud file path without scheme

        Return:
            whether path is valid
        """

        raise NotImplementedError
