"""AWS S3 pathlib support."""


import fnmatch
import importlib
import os
import re
from types import ModuleType
from typing import Any, ClassVar, Iterator, Optional

from cloudpath.base import base
from cloudpath import cloud


class S3Path(cloud.CloudPath):
    """AWS S3 path manager."""

    _boto3: ClassVar[ModuleType]
    _endpoint: ClassVar[Optional[str]]
    _resource: ClassVar[Any]
    _region: ClassVar[str]

    def _download_file(self, local_path: str, remote_path: str) -> bool:
        """Download file from remote storage bucket.

        Args:
            local_path: local system file path
            remote_path: remote storage file path

        Return:
            whether file download was successful
        """

        self._bucket.download_file(Filename=local_path, Key=remote_path)

        return True

    def exists(self) -> bool:
        """Check whether the path points to an existing file or directory.

        Return:
            whether path exists
        """

        self._load_bucket()
        key_path = os.path.join(*self._path.parts[2:])

        try:
            next(iter(self._bucket.objects.filter(Prefix=key_path)))
        except StopIteration:
            return False
        else:
            return True

    def _get_key_path(self) -> str:
        """Get S3 bucket key path.

        Return:
            S3 bucket key path
        """

        parts = self._path.parts[2:]
        if parts:
            return os.path.join(*parts)
        else:
            return ""

    def glob(self, pattern: str) -> Iterator["S3Path"]:
        """Glob the relative pattern in the path directory.

        Args:
            pattern: glob pattern

        Yields:
            paths matching the glob pattern
        """

        self._load_bucket()

        key_path = self._get_key_path()
        for obj in self._bucket.objects.filter(Prefix=key_path):
            path = f"/{obj.bucket_name}/{obj.key}"
            rel_path = os.path.relpath(path, self._path)

            if base.file_basename(rel_path) and fnmatch.fnmatchcase(rel_path, pattern):
                yield S3Path(path)

    def _load_bucket(self) -> None:
        """Find and set cloud storage bucket.

        Raises:
            FileNotFoundError: if unable to find storage bucket
        """

        if hasattr(self, "_bucket"):
            return

        for bucket in self._resource.buckets.all():
            if bucket.name == self._bucket_name:
                break
        else:
            raise FileNotFoundError(f"unable to access bucket {self._bucket_name}")

        self._bucket = bucket

    @classmethod
    def _load_resource(cls) -> None:
        """Import the boto3 library for handling AWS S3 paths.

        Dynamically imported so that users can use just PathFS without installing
        boto3.
        """

        if not hasattr(cls, "_boto3"):
            cls._boto3 = importlib.import_module("boto3")
            cls.settings(mutate=False)

        if not hasattr(cls, "_client"):
            cls._resource = cls._boto3.resource(  # type: ignore
                "s3", endpoint_url=cls._endpoint, use_ssl=True, region_name=cls._region
            )

    def rglob(self, pattern: str) -> Iterator["S3Path"]:
        """Recursively glob the relative pattern in the path directory.

        Args:
            pattern: glob pattern

        Yields:
            paths matching the glob pattern
        """

        self._load_bucket()

        key_path = self._get_key_path()
        for obj in self._bucket.objects.filter(Prefix=key_path):
            path = f"/{obj.bucket_name}/{obj.key}"
            if fnmatch.fnmatchcase(path.replace("key_path", ""), pattern):
                yield S3Path(path)

    @classmethod
    def scheme(cls) -> str:
        """Return cloud storage URI scheme.

        Return:
            URI scheme
        """

        return "s3"

    @classmethod
    def settings(
        cls, endpoint: Optional[str] = None, region: str = "us-west-2", mutate: bool = True
    ) -> None:
        """Configure general S3 settings.

        Args:
            endpoint: connection endpoint
            region: AWS region
            mutate: whether to mutate the settings if they already exist
        """

        if mutate:
            cls._endpoint = endpoint
            cls._region = region
        else:
            cls._endpoint = cls._endpoint if hasattr(cls, "_endpoint") else endpoint
            cls._region = cls._region if hasattr(cls, "_region") else region

    def _upload_file(self, local_path: str, remote_path: str) -> bool:
        """Upload file to remote storage bucket.

        Args:
            local_path: local system file path
            remote_path: remote storage file path

        Return:
            whether upload was successful
        """

        self._bucket.upload_file(
            Filename=local_path,
            Key=remote_path,
            ExtraArgs={
                "ContentType": "text/plain",
                "ServerSideEncryption": "AES256",
                "StorageClass": "STANDARD",
            },
        )

        return True

    @classmethod
    def valid_path(cls, path_name: str) -> bool:
        """Check if S3 path is valid.

        Args:
            full S3 file path without scheme

        Return:
            whether path is valid
        """

        # Naming requirements
        # https://docs.aws.amazon.com/awscloudtrail/latest/userguide/cloudtrail-s3-bucket-naming-requirements.html  # noqa: E501

        # TODO: disallow consecutive periods or dashes adjacent to periods

        regex = re.compile(
            """
            /  # path must contain
            [a-z0-9.-]{2,62}[a-z0-9.]  # S3 bucket naming requirements
            /.*
            """,
            re.VERBOSE,
        )

        return bool(regex.match(path_name))
